from pathlib import Path
import tempfile

import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import locan
from locan import LocData
from locan import ROOT_DIR
from locan.locan_io.locdata.io_locdata import load_txt_file
from locan.data.rois import RoiLegacy_0
from locan.render.utilities import _napari_shape_to_RoiRegion
from locan.data import metadata_pb2


# fixtures

@pytest.fixture()
def locdata():
    locdata_dict = {
        'position_x': [0, 1, 2, 3, 0, 1, 4, 5],
        'position_y': [0, 1, 2, 3, 1, 4, 5, 1],
        'position_z': [0, 1, 2, 3, 4, 4, 4, 5]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def points():
    return np.array([[0, 0], [0.5, 0.5], [100, 100], [1, 1], [1.1, 0.9]])


# tests

def test_Roi_2d(locdata_2d):
    roi = RoiLegacy_0(reference=locdata_2d, region_specs=((0, 0), 2, 2, 0), region_type='rectangle')
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is roi.region
    assert isinstance(roi.region, locan.RoiRegion)
    roi.region = locan.RoiRegion(region_specs=((0, 0), 2, 2, 0), region_type='rectangle')
    assert isinstance(roi.region, locan.RoiRegion)

    roi = RoiLegacy_0(reference=locdata_2d, region_specs=((0, 0), 4, 4, 0), region_type='rectangle',
              properties_for_roi=('position_y', 'frame'))
    new_dat = roi.locdata()
    assert len(new_dat) == 2
    assert new_dat.region is None

    locdata_empty = LocData()
    roi_empty = RoiLegacy_0(reference=locdata_empty, region_specs=((0, 0), 2, 1, 10), region_type='rectangle')
    empty_dat = roi_empty.locdata()
    assert len(empty_dat) == 0
    assert new_dat.region is None


def test_Roi_3d(locdata_3d):
    roi = RoiLegacy_0(reference=locdata_3d, region_specs=((0, 0), 2, 2, 0), region_type='rectangle')
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None
    assert isinstance(roi.region, locan.RoiRegion)
    roi.region = locan.RoiRegion(region_specs=((0, 0), 2, 1, 0), region_type='rectangle')
    assert isinstance(roi.region, locan.RoiRegion)

    roi = RoiLegacy_0(reference=locdata_3d, region_specs=((0, 0), 2, 3, 0), region_type='rectangle',
              properties_for_roi=('position_y', 'position_z'))
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None


def test_pickling_locdata_from_Roi(locdata_2d):
    import pickle
    roi = RoiLegacy_0(reference=locdata_2d, region_specs=((0, 0), 2, 1, 0), region_type='rectangle')
    new_dat = roi.locdata()

    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_locdata.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(new_dat, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            locdata = pickle.load(file)
        assert len(new_dat) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)


def test_Roi_io(locdata):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'roi.yaml'
        #file_path = ROOT_DIR / 'tests/test_data/roi.yaml'

        roi = RoiLegacy_0(reference=locdata, region_type='rectangle', region_specs=((0, 0), 2, 1, 10))
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new.reference is None

        roi = RoiLegacy_0(reference=dict(file_path=ROOT_DIR / 'tests/test_data/five_blobs.txt', file_type=1),
                  region_type='rectangle', region_specs=((0, 0), 2, 1, 10))
        assert isinstance(roi.reference, (metadata_pb2.Metadata, Path))
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new

        # test region specs with numpy floats
        roi = RoiLegacy_0(reference=dict(file_path=file_path, file_type=1),
                  region_type='rectangle',
                  region_specs=(np.array([0, 0], dtype=float), float(2), float(1), float(10))
                  )
        assert isinstance(roi.reference, metadata_pb2.Metadata)
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new

        meta_ = metadata_pb2.Metadata()
        meta_.file.path = str(file_path)
        meta_.file.type = 1
        locdata_2 = LocData.from_selection(locdata, meta=meta_)
        roi = RoiLegacy_0(reference=locdata_2,
                  region_type='rectangle', region_specs=((0, 0), 2, 1, 10))
        assert isinstance(roi.reference.meta, metadata_pb2.Metadata)
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new
        assert isinstance(roi_new.reference, metadata_pb2.Metadata)


def test_roi_locdata_from_file():
    locdata = load_txt_file(path=ROOT_DIR / 'tests/test_data/five_blobs_3D.txt')

    roi = RoiLegacy_0(reference=locdata, region_type='rectangle', region_specs=((100, 500), 200, 200, 10))
    dat = roi.locdata()
    assert(len(dat) == 10)

    roi = RoiLegacy_0(reference=locdata, region_type='rectangle', region_specs=((100, 100), 200, 200, 10),
              properties_for_roi=['position_x', 'position_z'])
    dat = roi.locdata()
    assert(len(dat) == 10)


def test_as_artist():
    roi_rectangle = RoiLegacy_0(reference=locdata, region_specs=((0, 0), 0.5, 0.5, 10), region_type='rectangle')
    roi_ellipse = RoiLegacy_0(reference=locdata, region_specs=((0.5, 0.5), 0.2, 0.3, 10), region_type='ellipse')
    roi_polygon = RoiLegacy_0(reference=locdata, region_specs=((0, 0.8), (0.1, 0.9), (0.6, 0.9), (0.7, 0.7), (0, 0.8)),
                      region_type='polygon')
    fig, ax = plt.subplots()
    ax.add_patch(roi_rectangle._region.as_artist())
    ax.add_patch(roi_ellipse._region.as_artist())
    ax.add_patch(roi_polygon._region.as_artist())
    # plt.show()

    plt.close('all')


def test__napari_shape_to_RoiRegion():
    # rectangle
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_RoiRegion(vertices, bin_edges, 'rectangle')
    for a, b in zip(region.region_specs, ((0.0, 2.0), 25.0, 3.0999999999999996, 0)):
        assert a == b

    # ellipse
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_RoiRegion(vertices, bin_edges, 'ellipse')
    for a, b in zip(region.region_specs, ((12.5, 3.55), 25.0, 3.0999999999999996, 0)):
        assert a == b

    # polygon
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_RoiRegion(vertices, bin_edges, 'polygon')
    assert np.array_equal(region.region_specs, np.array([[0, 2], [25, 2], [25, 5.1], [0, 5.1], [0, 2]]))



"""

Alpha shape

This module provides methods for computing the alpha complex and specific alpha shapes
together with related properties for LocData objects.

Alpha shape is a hull object that
defines a group of localizations bordering a concave hull (which does not have to be connected and might have
holes) [1]_. It depends on a single parameter `alpha` (here defined to be a distance with a unit equal to the coordinate
units of the given localization data). For `alpha` approaching infinity the alpha shape is equal to the convex hull.

In this context we call an alpha-complex the subgroup of simplexes of a Delaunay triangulation that is computed
according to Edelsbrunner algorithm.  All localizations that lie on the boundary of the alpha-complex make up the
alpha shape.

Internally we also work with an alpha-independent representation of the alpha-complex that allows efficient computation
of alpha shapes for arbitrary alpha values.

Simplices are classified as exterior (not part of the alpha complex), interior (not part of the boundary),
regular (part of the boundary but not singular), and singular (part of the boundary but all simplices of higher
dimension they are incident to are exterior).

References
----------
.. [1] H. Edelsbrunner and E. P. Mücke, Three-dimensional alpha shapes. ACM Trans. Graph. 13(1):43-72, 1994.
"""
import warnings

import numpy as np
from scipy.spatial import Delaunay
import networkx as nx

from locan.data.region import Region2D, Polygon, MultiPolygon
from locan.data.region_utils import regions_union
from locan.data.hulls.alpha_shape_2d import _circumcircle, _half_distance


__all__ = ['AlphaComplex', 'AlphaShape']


def _k_simplex_index_list(d, k):
    """Given a d-simplex with d being 2 or 3, indexes are provided for all k-simplices with k<d
    that are part of the d-simplex."""
    if d == 2:
        if k == 0:
            return 0, 1, 2
        elif k == 1:
            return (0, 1), (1, 2), (2, 0)
        else:
            raise ValueError
    elif d == 3:
        if k == 0:
            return 0, 1, 2, 3
        elif k == 1:
            return (0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)
        elif k == 2:
            return (0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)
        else:
            raise ValueError
    else:
        raise ValueError


def _k_simplex_neighbor_index_list(d, k):
    """Given a d-simplex with d being 2 or 3, indexes for neighbor simplexes in the scipy.Delaunay triangulation
    are provided for all k-simplices with k<d that are part of the d-simplex."""
    if d == 2 and k == 1:
        return 2, 0, 1
    elif d == 3 and k == 2:
        return 3, 2, 1, 0
    else:
        raise ValueError


def _get_k_simplices(simplex, k=1):
    """ Function to extract k-simplices (e.g. edges) from a d-simplex (e.g. triangle; d>k)."""
    d = len(simplex) - 1
    k_simplices = ([simplex[indx] for indx in indices] for indices in _k_simplex_index_list(d, k))
    return k_simplices


class AlphaComplex:
    """
    Class for an alpha-independent representation of the alpha complex of the given points.

    Here the alpha complex is the simplicial subcomplex of the Denlaunay triangulation together with the intervals
    defining simplex membership for an alpha complex for a specific `alpha`.

    Parameters
    ----------
    points : array-like with shape (npoints, ndim)
        Coordinates of input points.
    delaunay : scipy.spatial.Delaunay
        Object with attribute `simplices` specifying a list of indices in the array of points that define the
        simplexes in the Delaunay triangulation.
        Also an attribute `neighbor` is required that specifies indices of neighboring simplices.
        If None, scipy.stat.Delaunay(points) is computed.

    Attributes
    ----------
    lines : list of tuple with shape (n_lines, 3)
        1-simplices (lines) that represent a simplicial subcomplex of the Delaunay triangulation with intervals.
    triangles : list of tuple with shape (n_triangles, 3)
        2-simplices (triangles) that represent a simplicial subcomplex of the Delaunay triangulation with intervals.
    tetrahedrons : list of tuple with shape (n_tetrahedrons, 3)
        3-simplices (tetrahedrons) that represent a simplicial subcomplex of the Delaunay triangulation with intervals.
    dimension : int
        Spatial dimension of the hull.
    delaunay_triangulation : scipy.spatial.Delaunay
        Object with attribute `simplices` specifying a list of indices in the array of points that define the
        simplexes in the Delaunay triangulation.
        Also an attribute `neighbor` is required that specifies indices of neighboring simplices.
    """

    def __init__(self, points, delaunay=None):
        self.points = np.asarray(points)

        if np.size(self.points) == 0:
            self.dimension = None
            self.delaunay_triangulation = None
            self.lines = []
            self.triangles = []

        elif np.size(self.points) <= 4 and delaunay is None:
            warnings.warn("Not enough points to construct initial simplex (need 4)")
            self.dimension = None
            self.delaunay_triangulation = None
            self.lines = []
            self.triangles = []

        else:
            self.dimension = np.shape(self.points)[1]
            self.delaunay_triangulation = Delaunay(self.points) if delaunay is None else delaunay

            if self.dimension == 2:
                self.lines, self.triangles = self._compute_2d()
            elif self.dimension == 3:
                self.lines = []
                self.triangles = []
                raise NotImplementedError
            else:
                raise ValueError(f'There is no algorithm available for points with dimension {self.dimension}.')

    def _compute_2d(self):
        """ Compute the alpha complex for 2d data."""
        n_simplices = len(self.delaunay_triangulation.simplices)

        # circumference for d-simplexes
        circumcircle_radii = np.zeros(n_simplices)
        for n, simplex in enumerate(self.delaunay_triangulation.simplices):
            _, circumcircle_radii[n] = _circumcircle(self.points, simplex)

        alpha_complex_lines = []
        alpha_complex_triangles = []
        for n, (simplex, neighbors, circumcircle_radius) in enumerate(zip(self.delaunay_triangulation.simplices,
                                                                     self.delaunay_triangulation.neighbors,
                                                                     circumcircle_radii)):
            interval_a_list = []
            interval_b_list = []
            interval_c_list = []
            for k_simplex, neighbor_index in zip(_get_k_simplices(simplex, k=1),
                                                 _k_simplex_neighbor_index_list(d=2, k=1)):
                neighbor = neighbors[neighbor_index]
                circumcircle_radius_neigh = np.inf if neighbor == -1 else circumcircle_radii[neighbor]
                interval_b, interval_c = sorted((circumcircle_radius, circumcircle_radius_neigh))
                interval_a = _half_distance(self.points[k_simplex])  # distance between vertices / 2

                # named tuple Simplices(vertices, interval_a, interval_b, interval_c)
                alpha_complex_lines.append((tuple(sorted(k_simplex)), interval_a, interval_b, interval_c))
                # use tuple(sorted()) together with set to get rid of duplicate simplices

                interval_a_list.append(interval_a)
                interval_b_list.append(interval_b)
                interval_c_list.append(interval_c)

            # if alpha is large enough such that ALL k-simplices are singular, regular, or interior,
            # the d-simplex is a singular, regular or interior part of the alpha shape
            alpha_complex_triangles.append((n, max(interval_a_list), max(interval_b_list), max(interval_c_list)))
        alpha_complex_lines = set(alpha_complex_lines)
        return alpha_complex_lines, alpha_complex_triangles

    def get_alpha_complex_lines(self, alpha, type='all'):
        """
        Simplicial subcomplex (lines) of the Delaunay triangulation for the specific alpha complex
        for the given `alpha`.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.
        type : str
            Type of alpha complex edges to be included in the graph.
            One of 'all', 'regular', 'singular', 'interior', 'exterior'.

        Returns
        -------
        list of list of int
            The indices to specific points in `self.points`.
        """
        if np.size(self.lines) == 0:
            return []

        if type == 'exterior':
            return [list(ac[0]) for ac in self.lines if ac[1] > alpha]
        elif type == 'all':
            return [list(ac[0]) for ac in self.lines if ac[1] <= alpha]
        elif type == 'singular':
            return [list(ac[0]) for ac in self.lines if ac[1] <= alpha < ac[2]]
        elif type == 'regular':
            return [list(ac[0]) for ac in self.lines if ac[2] <= alpha < ac[3]]
        elif type == 'interior':
            return [list(ac[0]) for ac in self.lines if ac[3] <= alpha]
        else:
            raise AttributeError(f'Parameter type: {type} is not valid.')
        # a list of lists and not of tuples should be returned since the return value will be used for indexing arrays.

    def get_alpha_complex_triangles(self, alpha, type='all'):
        """
        Simplicial subcomplex (triangles) of the Delaunay triangulation for the specific alpha complex
        for the given `alpha`.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.
        type : str
            Type of alpha complex edges to be included in the graph.
            One of 'all', 'regular', 'singular', 'interior', 'exterior'.

        Returns
        -------
        list of list of int
            The indices to specific d-simplices in `self.delaunay_triangulation.simplices`.
        """
        if np.size(self.triangles) == 0:
            return []

        if type == 'exterior':
            return [ac[0] for ac in self.triangles if ac[1] > alpha]
        elif type == 'all':
            return [ac[0] for ac in self.triangles if ac[1] <= alpha]
        elif type == 'singular':
            return [ac[0] for ac in self.triangles if ac[1] <= alpha < ac[2]]
        elif type == 'regular':
            return [ac[0] for ac in self.triangles if ac[2] <= alpha < ac[3]]
        elif type == 'interior':
            return [ac[0] for ac in self.triangles if ac[3] <= alpha]
        else:
            raise AttributeError(f'Parameter type: {type} is not valid.')
        # a list of lists and not of tuples should be returned since the return value will be used for indexing arrays.

    def graph_from_lines(self, alpha, type='all'):
        """
        Return networkx Graph object with nodes and edges from selected lines.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.
        type : str
            Type of alpha complex edges to be included in the graph.
            One of 'all', 'regular', 'singular', 'interior', 'exterior'.

        Returns
        -------
        networkx.Graph
        """
        G = nx.Graph()

        if np.size(self.lines) == 0:
            return G

        # positions = {i: tuple(point) for i, point in enumerate(self.points)}  # positions for all nodes
        # G.add_nodes_from(positions)
        ac_simplices = self.get_alpha_complex_lines(alpha, type)
        G.add_edges_from(ac_simplices, type=type)
        return G

    def graph_from_triangles(self, alpha, type='all'):
        """
        Return networkx Graph object with nodes and edges from selected triangles.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.
        type : str
            Type of alpha complex edges to be included in the graph.
            One of 'all', 'regular', 'singular', 'interior', 'exterior'.

        Returns
        -------
        networkx.Graph
        """
        G = nx.MultiGraph()

        if np.size(self.triangles) == 0:
            return G

        # positions = {i: tuple(point) for i, point in enumerate(self.points)}  # positions for all nodes
        # G.add_nodes_from(positions)
        triangles = self.get_alpha_complex_triangles(alpha=alpha, type=type)
        for triangle, vertices in zip(triangles, self.delaunay_triangulation.simplices[triangles]):
            edges = [vertices[[n, m]] for n, m in [(0, 1), (1, 2), (2, 0)]]
            G.add_edges_from(edges, triangle=triangle)
        return G

    def alpha_shape(self, alpha):
        """
        Return the unique alpha shape for `alpha`.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.

        Returns
        -------
        float
        """
        return AlphaShape(points=self.points, alpha=alpha,
                          alpha_complex=self, delaunay=self.delaunay_triangulation)

    def optimal_alpha(self):
        """
        Find the minimum alpha value for which all points belong to the alpha shape,
        in other words, no edges of the Delaunay triangulation are exterior.

        Returns
        -------
        float
        """
        if np.size(self.lines) == 0:
            return None
        else:
            return np.max([ac[1] for ac in self.lines])

    def alphas(self):
        """
        Return alpha values at which the corresponding alpha shape changes.

        Returns
        -------
        numpy.ndarray
        """
        if np.size(self.lines) == 0:
            return np.array([])
        else:
            return np.unique([ac[1:] for ac in self.lines])


class AlphaShape:
    """
    Class for the alpha shape of points for a specific alpha value.

    Here the alpha complex is the simplicial subcomplex of the Denlaunay triangulation for a given alpha value.
    The alhpa shape is the union of all simplexes of the alpha complex, specified by the boundary points of the
    alpha complex.

    In order to update an existing AlphaShape object to a different `alpha` reset AlphaShape.alpha.

    Parameters
    ----------
    alpha : float
        Alpha parameter specifying a unique alpha complex.
    points : array-like with shape (npoints, ndim), None
        Coordinates of input points. Either `points` or `alpha_complex` have to be specified but not both.
    alpha_complex : AlphaComplex, None
        The unfiltered alpha complex with computed interval values.
    delaunay : scipy.spatial.Delaunay, None
        Object with attribute `simplices` specifying a list of indices in the array of points that define the
        simplexes in the Delaunay triangulation.
        Also an attribute `neighbor` is required that specifies indices of neighboring simplices.
        If None, scipy.stat.Delaunay(points) is computed.

    Attributes
    ----------
    alpha_complex : AlphaComplex
        The unfiltered alpha complex with computed interval values.
    alpha_shape : numpy.ndarray
        The list of k-simplices (edges) from the alpha complex that make up the alpha shape.
        Or: Simplicial subcomplex of the Delaunay triangulation with regular simplices from the alpha complex.
    region : Region
        Region object.
    connected_components : list of Region
        Connected components, i.e. a list of the individual unconnected polygons that together make up the alpha shape.
    dimension : int
        Spatial dimension of the hull as determined from the dimension of `points`
    vertices : array of (2,) tuples
        Coordinates of points that make up the hull (regular alpha_shape line-simplices).
    vertex_indices : list of int
        Indices identifying a polygon of all points that make up the hull (regular alpha_shape line-simplices).
    vertices_alpha_shape : array of (2,) tuples
        Coordinates of points that make up the interior and boundary of the hull
        (regular, singular and interior alpha_shape line-simplices).
    vertex_alpha_shape_indices : list of int
        Indices to all points that make up the interior and boundary of the hull.
        (regular, singular and interior alpha_shape line-simplices).
    vertices_connected_components_indices : list of list
        Indices to the points for each connected component of the alpha shape.
    n_points_on_boundary : float
        The number of points on the hull (regular and singular alpha_shape simplices).
    n_points_on_boundary_rel : float
        The number of points on the hull (regular and singular alpha_shape simplices)
        relative to all alpha_shape points.
    n_points_alpha_shape : int
        Absolute number of points that are part of the alpha_shape
        (regular, singular and interior alpha_shape simplices).
    n_points_alpha_shape_rel : int
        Absolute number of points that are part of the alpha_shape relative to all input points
        (regular, singular and interior alpha_shape simplices).
    region_measure : float
        Hull measure, i.e. area or volume.
    subregion_measure : float
        Measure of the sub-dimensional region, i.e. circumference or surface.
    """

    def __init__(self, alpha, points=None, alpha_complex=None, delaunay=None):
        if alpha_complex is None:
            self.alpha_complex = AlphaComplex(points, delaunay)
        else:
            self.alpha_complex = alpha_complex

        self.points = self.alpha_complex.points
        self.dimension = self.alpha_complex.dimension
        self._alpha = None
        self._region = None
        self._connected_components = None
        self._vertices_connected_components_indices = None

        self.alpha = alpha

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha
        self._region = None
        self._connected_components = None
        self._vertices_connected_components_indices = None

        self.n_points_alpha_shape = self.alpha_complex.graph_from_lines(alpha=self.alpha, type='all'). \
            number_of_nodes()
        self.n_points_on_boundary = self.alpha_complex.graph_from_lines(alpha=self.alpha, type='regular'). \
            number_of_nodes()
        try:
            self.n_points_on_boundary_rel = self.n_points_on_boundary / self.n_points_alpha_shape
        except ZeroDivisionError:
            self.n_points_on_boundary_rel = float('nan')

        try:
            self.n_points_alpha_shape_rel = self.n_points_alpha_shape / len(self.points)
        except ZeroDivisionError:
            self.n_points_alpha_shape_rel = float('nan')

        self.region_measure = self.region.region_measure
        self.subregion_measure = self.region.subregion_measure

    @property
    def region(self):
        if self.dimension == 3:
            raise NotImplementedError('Region for 3D data has not yet been implemented.')

        if self._region is None:
            triangles = self.alpha_complex.get_alpha_complex_triangles(alpha=self.alpha)
            if not triangles:
                vertices = []
            else:
                vertices = self.alpha_complex.delaunay_triangulation.simplices[triangles]
            self._region = regions_union([Polygon(pts) for pts in self.points[vertices]])
        return self._region

    @property
    def connected_components(self):
        if self._connected_components is None:
            self._vertices_connected_components_indices = []
            self._connected_components = []
            graph = self.alpha_complex.graph_from_triangles(alpha=self.alpha)
            for cc in nx.connected_components(graph):
                subgraph = graph.subgraph(cc)
                self._vertices_connected_components_indices.append(list(subgraph.nodes))
                triangles = list({edge[2] for edge in subgraph.edges.data('triangle')})
                vertices = self.alpha_complex.delaunay_triangulation.simplices[triangles]
                region = regions_union([Polygon(pts) for pts in self.points[vertices]])
                self._connected_components.append(region)
        return self._connected_components

    @property
    def vertices_connected_components_indices(self):
        _ = self.connected_components  # trigger computation of _vertices_connected_components_indices
        return self._vertices_connected_components_indices

    @property
    def alpha_shape(self):
        return self.alpha_complex.get_alpha_complex_lines(self.alpha, type='all')

    @property
    def vertices(self):
        return self.points[self.vertex_indices]

    @property
    def vertex_indices(self):
        array = self.alpha_complex.get_alpha_complex_lines(self.alpha, type='regular')
        return np.unique(array).tolist()

    @property
    def vertices_alpha_shape(self):
        return self.points[self.vertex_alpha_shape_indices]

    @property
    def vertex_alpha_shape_indices(self):
        array_r = self.alpha_complex.get_alpha_complex_lines(self.alpha, type='regular')
        array_s = self.alpha_complex.get_alpha_complex_lines(self.alpha, type='singular')
        array_i = self.alpha_complex.get_alpha_complex_lines(self.alpha, type='interior')
        return np.unique(array_r + array_s + array_i).tolist()



import numpy as np

from locan.render.utilities import _napari_shape_to_region


def test__napari_shape_to_region():
    # rectangle
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_region(vertices, bin_edges, 'rectangle')
    assert repr(region) == 'Rectangle((0.0, 2.0), 31.0, 2.5, 0)'

    # ellipse
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_region(vertices, bin_edges, 'ellipse')
    assert repr(region) == 'Ellipse((15.5, 3.25), 31.0, 2.5, 0)'

    # polygon
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    bin_edges = np.array([[0, 10, 20], [2, 3, 4, 5]], dtype=object)
    region = _napari_shape_to_region(vertices, bin_edges, 'polygon')
    assert repr(region) == 'Polygon([[0.0, 2.0], [0.0, 4.5], [31.0, 4.5], [31.0, 2.0], [0.0, 2.0]])'



"""

Synthetic data

This module provides functions to simulate localization and other data that can be used for testing and development.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   simulate_locdata
   simulate_drift

"""
from .simulate_locdata import *
from .simulate_drift import *


__all__ = []
__all__.extend(simulate_locdata.__all__)
__all__.extend(simulate_drift.__all__)


"""

File input/output for localization data in Elyra files.

"""
from __future__ import annotations
import io
import logging

import pandas as pd

from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, open_path_or_file_like, convert_property_names


__all__ = ['load_Elyra_header', 'load_Elyra_file']

logger = logging.getLogger(__name__)


def _read_Elyra_header(file):
    """
    Read xml header from a Zeiss Elyra single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        A rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    header = file.readline().split('\n')[0]

    # list identifiers
    identifiers = header.split("\t")

    column_keys = convert_property_names(properties=identifiers, property_mapping=locan.constants.ELYRA_KEYS)

    return column_keys


def load_Elyra_header(path):
    """
    Load xml header from a Zeiss Elyra single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """

    with open_path_or_file_like(path, encoding='latin-1') as file:
        return _read_Elyra_header(file)


def load_Elyra_file(path, nrows=None, convert=True, **kwargs):
    """
    Load data from a rapidSTORM single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.

    Note
    ----
    Data is loaded with encoding = 'latin-1' and only data before the first NUL character is returned.
    Additional information appended at the end of the file is thus ignored.
    """
    with open_path_or_file_like(path, encoding='latin-1') as file:
        columns = _read_Elyra_header(file)
        string = file.read()
        # remove metadata following nul byte
        string = string.split('\x00')[0]

        stream = io.StringIO(string)
        dataframe = pd.read_csv(stream, sep="\t", skiprows=0, nrows=nrows, names=columns, **kwargs)

    if convert:
        dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.ELYRA
    dat.meta.file.path = str(path)

    del dat.meta.history[:]
    dat.meta.history.add(name='load_Elyra_file', parameter='path={}, nrows={}'.format(path, nrows))

    return dat



"""

Compute localizations per frame.

"""
import logging
from dataclasses import dataclass

import numpy as np
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt

from locan.analysis.analysis_base import _Analysis


__all__ = ['LocalizationsPerFrame']

logger = logging.getLogger(__name__)


#### The algorithms

def _localizations_per_frame(locdata, norm=None, time_delta="integration_time", resample=None, **kwargs):
    """
    Compute localizations per frame.

    Parameters
    ----------
    locdata : LocData or array-like
       Points in time: either localization data that contains a column `frame` or an array with time points.
    norm : int, float, str, None
        Normalization factor that can be None, a number, or another property in `locdata`.
    time_delta : int, float, str, None
        Time per frame or "integration_time" for which the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample : DateOffset, Timedelta or str
        Parameter for :func:`pandas.Series.resample`: The offset string or object representing target conversion.
    kwargs : dict
        Other parameters passed to :func:`pandas.Series.resample`.


    Returns:
    --------
    pandas.Series
    """
    # normalization
    if norm is None:
        normalization_factor = 1
        series_name = 'n_localizations'
    elif isinstance(norm, str):
        normalization_factor = locdata.properties[norm]
        series_name = f'n_localizations / ' + norm
    elif isinstance(norm, (int, float)):
        normalization_factor = norm
        series_name = f'n_localizations / {norm}'
    else:
        raise TypeError('normalization should be None, a number or a valid property name.')

    try:
        frames_ = locdata.data.frame.astype(int)
    except AttributeError:
        frames_ = np.asarray(locdata)

    frames, frame_counts = np.unique(frames_, return_counts=True)
    series = pd.Series(frame_counts, index=frames, dtype=float) / normalization_factor
    series.index.name = 'frame'
    series.name = series_name

    if time_delta is None:
        pass
    elif isinstance(time_delta, (int, float)):
        series.index = pd.to_timedelta(frames * time_delta, unit="s")
        series.index.name = 'time'
        series.name = series.name + " / s"
    elif time_delta == "integration_time":
        try:
            time_delta = locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time.ToSeconds()
            series.index = pd.to_timedelta(frames * time_delta, unit="s")
            series.index.name = 'time'
            series.name = series.name + " / s"
        except (IndexError, AttributeError):
            logger.warning("integration_time not available in locdata.meta - frames used instead.")
    else:
        raise ValueError("The input for time_delta is not implemented.")

    if resample is not None:
        series = series.resample(resample, **kwargs).sum()

    return series


# The specific analysis classes

@dataclass(repr=False)
class _Results:
    time_series: pd.Series

    def accumulation_time(self, fraction=0.5) -> int:
        normalized_cumulative_time_trace = self.time_series.cumsum() / self.time_series.sum()
        accumulation_time = normalized_cumulative_time_trace.gt(fraction).idxmax()
        return accumulation_time


class LocalizationsPerFrame(_Analysis):
    """
    Compute localizations per frame.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    norm : int, float, str, None
        Normalization factor that can be None, a number, or another property in `locdata`.
    time_delta : int, float, str, None
        Time per frame or "integration_time" for which the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample : DateOffset, Timedelta or str
        Parameter for :func:`pandas.Series.resample`: The offset string or object representing target conversion.
    kwargs : dict
        Other parameters passed to :func:`pandas.Series.resample`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.Series
        Computed results.
    distribution_statistics : Distribution_fits object, None
        Distribution parameters derived from MLE fitting of results.
    """
    count = 0

    def __init__(self, meta=None, norm=None, time_delta="integration_time", resample=None, **kwargs):
        super().__init__(meta=meta, norm=norm, time_delta=time_delta, resample=resample, **kwargs)
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData or array-like
           Points in time: either localization data that contains a column `frame` or an array with time points.

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            self.distribution_statistics = None
            self.results = None
            return self

        self.distribution_statistics = None
        self.results = _Results(
            time_series=_localizations_per_frame(locdata=locdata, **self.parameter)
        )
        return self

    def fit_distributions(self, **kwargs):
        """
        Fit probability density functions to the distributions of `loc_property` values in the results
        using MLE (scipy.stats).

        Parameters
        ----------
        loc_property : str
            The LocData property for which to fit an appropriate distribution; if None all plots are shown.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(**kwargs)
        else:
            logger.warning('No results available to fit.')

    def plot(self, ax=None, window=1, cumulative=False, normalize=False, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the running average of results over window size.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        window: int
            Window for running average that is applied before plotting.
        cumulative : bool
            Plot the cumulated results if true.
        normalize : bool
            Normalize cumulative plot to the last value
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        series = self.results.time_series

        # prepare plot
        if cumulative and normalize:
            _results = series.cumsum() / series.sum()
        elif cumulative and not normalize:
            _results = series.cumsum()
        elif not cumulative:
            _results = series
        else:
            return ax

        _results.rolling(window=window, center=True).mean().plot(ax=ax, **kwargs)

        ax.set(title=f'Localizations per Frame\n (window={window})',
               xlabel=series.index.name,
               ylabel=f'{series.name} (cumulative)' if cumulative else series.name
               )

        return ax

    def hist(self, ax=None, fit=True, bins='auto', **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing hist(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        bins : float
            Bin specifications (passed to matplotlib.hist).
        fit: Bool
            Flag indicating if distributions fit are shown.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        series = self.results.time_series

        ax.hist(series.values, bins=bins, **dict(dict(density=True, log=False), **kwargs))
        ax.set(title = 'Localizations per Frame',
               xlabel = series.name,
               ylabel = 'PDF'
               )

        if fit:
            if isinstance(self.distribution_statistics, _DistributionFits):
                self.distribution_statistics.plot(ax=ax)
            else:
                self.fit_distributions()
                self.distribution_statistics.plot(ax=ax)

        return ax

# todo: add fit function

class _DistributionFits:
    """
    Handle for distribution fits.

    It holds the statistical parameters derived by fitting the result distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationPrecision object
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.
    loc_property : str
        The LocData property for which to fit an appropriate distribution
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    parameters :
    """
    def __init__(self, analysis_class):
        self.analysis_class = analysis_class
        self.loc_property = self.analysis_class.results.time_series.name
        self.distribution = None
        self.parameters = []

    def fit(self, distribution=stats.norm, **kwargs):
        """
        Fit scipy.stats.distribution to analysis_class.results[loc_property].

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        kwargs : dict
            Other parameters are passed to the `scipy.stat.distribution.fit()` function.
        """
        if self.analysis_class.results is None:
            return

        self.distribution = distribution

        loc, scale = self.distribution.fit(self.analysis_class.results.time_series.values, **kwargs)
        self.parameters.extend([self.loc_property + '_center', self.loc_property + '_sigma'])
        setattr(self, self.loc_property + '_center', loc)
        setattr(self, self.loc_property + '_sigma', scale)


    def plot(self, ax=None, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the probability distribution functions of fitted
        results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        _center, _sigma = self.parameter_dict().values()

        x_values = np.linspace(self.distribution.ppf(0.001, loc=_center, scale=_sigma),
                               self.distribution.ppf(0.999, loc=_center, scale=_sigma), 100)
        ax.plot(x_values, self.distribution.pdf(x_values, loc=_center, scale=_sigma), 'r-',
                **dict(dict(lw=3, alpha=0.6, label=str(self.distribution.name) + ' pdf'), **kwargs))
        return ax

    def parameter_dict(self):
        """ Dictionary of fitted parameters. """
        if self.parameters is None:
            return None
        else:
            return {k: self.__dict__[k] for k in self.parameters}



import numpy as np
import pytest

from locan.render.transform import Transform
from locan import Trafo, adjust_contrast, HistogramEqualization


def test_Trafo():
    assert list(Trafo)


def test_HistogramEqualization():
    img = np.array((1, 2, 3, 9), dtype=np.float64)
    norm = HistogramEqualization(power=1, n_bins=4)
    print(norm)
    assert isinstance(norm, Transform)
    new_img = norm(img)
    assert np.allclose(new_img, [0.5, 0.5, 0.625, 1.])

    norm = HistogramEqualization(power=1, n_bins=4, vmax=3)
    assert isinstance(norm, Transform)
    new_img = norm(img)
    assert np.allclose(new_img, [0.25, 0.375, 1., 1.])

    norm = HistogramEqualization(power=1, n_bins=4, vmin=3)
    assert isinstance(norm, Transform)
    new_img = norm(img)
    assert np.allclose(new_img, [0.75, 0.75, 0.75, 1.])

    norm = HistogramEqualization(power=1, n_bins=3, mask=img > 1)
    assert isinstance(norm, Transform)
    new_img = norm(img)
    assert np.allclose(new_img, [0.66666667, 0.66666667, 0.66666667, 1.])

    norm = HistogramEqualization(power=3, n_bins=3)
    new_img = norm(img)
    assert np.allclose(new_img, [0.96428571, 0.96428571, 0.96428571, 1.])

    norm = HistogramEqualization(power=0.3, n_bins=3)
    new_img = norm(img)
    assert np.allclose(new_img, [0.58165808, 0.58165808, 0.58165808, 1.])


def test_adjust_contrast():
    img = np.array((1, 2, 3, 9), dtype=np.float64)
    new_img = adjust_contrast(img, rescale=(50, 100))
    assert np.array_equal(new_img, np.array((0, 0, 0, 1)))

    img = np.array((1, 2, 3, 9), dtype=np.uint8)
    new_img = adjust_contrast(img, rescale=None)
    assert np.array_equal(img, new_img)

    new_img = adjust_contrast(img, rescale=False)
    assert np.array_equal(img, new_img)

    new_img = adjust_contrast(img, rescale='equal')
    assert max(new_img) == 1

    new_img = adjust_contrast(img, rescale=True)
    # assert np.array_equal(new_img, np.array((0, 31, 63, 255)))
    assert np.array_equal(new_img, np.array((0, 0.125, 0.25, 1.)))

    new_img = adjust_contrast(img, rescale=(0, 50))
    assert np.array_equal(new_img, np.array((0, 63, 127, 255)))

    new_img = adjust_contrast(img, out_range=(0, 10))
    assert np.array_equal(new_img, np.array((0, 1.25, 2.5, 10)))

    new_img = adjust_contrast(img * 1., rescale=True)
    assert np.array_equal(new_img, np.array((0, 0.125, 0.25, 1.)))

    new_img = adjust_contrast(img, rescale='unity')
    assert np.array_equal(new_img, np.array((0, 0.125, 0.25, 1.)))

    norm = HistogramEqualization(power=1, n_bins=4)
    new_img = adjust_contrast(img, rescale=norm)
    assert np.allclose(new_img, [0.5, 0.5, 0.625, 1.])


def test_transform_new():
    img = np.array((2, 2.5, 3, 9), dtype=np.float64)
    img_with_zeros = np.array((2, 0, 2.5, 3, 9), dtype=np.float64)

    # 0 No transformation
    new_img = adjust_contrast(img, rescale=0)
    assert new_img is img
    new_img = adjust_contrast(img, rescale=None)
    assert new_img is img
    new_img = adjust_contrast(img, rescale=False)
    assert new_img is img
    new_img = adjust_contrast(img, rescale=Trafo.NONE)
    assert new_img is img

    # 1) standardize: rescale (min, max) to (0, 1)
    new_img = adjust_contrast(img, rescale=1)
    assert new_img.dtype == float
    assert (new_img.min(), new_img.max()) == (0, 1)

    new_img = adjust_contrast(img, rescale=Trafo.STANDARDIZE)
    assert new_img.dtype == float
    assert (new_img.min(), new_img.max()) == (0, 1)

    new_img = adjust_contrast(img, rescale="standardize")
    assert new_img.dtype == float
    assert (new_img.min(), new_img.max()) == (0, 1)

    # 2) standardize_uint8: rescale (min, max) to (0, 255)
    new_img = adjust_contrast(img, rescale=Trafo.STANDARDIZE_UINT8)
    assert new_img.dtype == np.uint8
    assert (new_img.min(), new_img.max()) == (0, 255)

    # 3) zero: (0, max) to (0, 1)
    new_img = adjust_contrast(img, rescale=Trafo.ZERO)
    assert (new_img.min(), new_img.max()) == (pytest.approx(0.2222222222222222), 1)

    # 4) zero_uint8: (0, max) to (0, 255)
    new_img = adjust_contrast(img, rescale=Trafo.ZERO_UINT8)
    assert new_img.dtype == np.uint8
    assert (new_img.min(), new_img.max()) == (56, 255)

    # 5) equalize: equalize histogram for all values>0
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE)
    assert new_img.dtype == np.float64
    assert np.allclose(new_img, [0.25, 0., 0.48611111, 0.70833333, 1.])

    # 6) equalize_uint8: equalize histogram for all values>0 within (0, 255)
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_UINT8)
    assert new_img.dtype == np.uint8
    assert np.allclose(new_img, [63, 0, 123, 180, 255])

    # 7) equalize_all: equalize histogram
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_ALL)
    assert new_img.dtype == np.float64
    assert np.allclose(new_img, [0.4, 0.2, 0.58888889, 0.76666667, 1.])

    # 8) equalize_all_uint8: equalize histogram within (0, 255)
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_ALL_UINT8)
    assert new_img.dtype == np.uint8
    assert np.allclose(new_img, [102, 51, 150, 195, 255])

    # 9) equalize_0p3: equalize histogram with factor=0.3 for all values>0
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_0P3)
    assert new_img.dtype == np.float64
    assert np.allclose(new_img, [0.25, 0., 0.48611111, 0.70833333, 1.])
    # result is the same as for EQUALIZE due to the large number of bins

    # 10) equalize_0p3_uint8: equalize histogram with factor=0.3 for all values>0 within (0, 255)
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_0P3_UINT8)
    assert new_img.dtype == np.uint8
    assert np.allclose(new_img, [63, 0, 123, 180, 255])

    # 11) equalize_0p3_all: equalize histogram with factor=0.3
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_0P3_ALL)
    assert new_img.dtype == np.float64
    assert np.allclose(new_img, [0.4, 0.2, 0.58888889, 0.76666667, 1.])

    # 12) equalize_0p3_all_uint8: equalize histogram with factor=0.3 within (0, 255)
    new_img = adjust_contrast(img_with_zeros, rescale=Trafo.EQUALIZE_0P3_ALL_UINT8)
    assert new_img.dtype == np.uint8
    assert np.allclose(new_img, [102, 51, 150, 195, 255])



"""
Example for a specialized analysis class.

It includes two algorithms for specific analysis routines.
And it provides standard interface functions modified for the specific analysis routine like report.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats

from locan.analysis.analysis_base import _Analysis

#
#### The algorithms
#
# First there is an algorithm to compute something given a set of point coordinates.
# Second there might be an alternative algorithm doing the same thing.
# Both can be tested with some simple point data.
#
# The user will not have to use this particular function directly.
#


def _algorithm_1(data=None, limits=(0, 10)):
    """ Provides a list of data values. data would be input data that is currently not used."""
    results = [i for i in range(*limits)]  # some complicated algorithm
    results = pd.DataFrame.from_dict(dict(a=results))
    return results


def _algorithm_2(data=None, n_sample=100, seed=None):
    """ Provides random normal distributed data. data would be input data that is currently not used."""
    np.random.seed(seed)
    dict_ = {'a': np.random.normal(size=n_sample),
             'b': np.random.normal(size=n_sample)}
    results = pd.DataFrame.from_dict(dict_)
    return results

#
##### The base analysis class
#
# Now we want a class implementing this algorithm to be used with locdata. Also the results from this algorithm should
# be reused in some visual representation that is specific for this analysis routine.
# Therefore we have a class that holds results and organizes metadata and provides the specific plotting routine.
#

# class: analysis.analysis_base._Analysis


#
# This specific analysis classes inherit from _Analysis.
#
# The classes for each particular algorithm are defined as:
#

class AnalysisExampleAlgorithm_1(_Analysis):
    """
    Example for an analysis class implementing algorithm_1.
    Compute some data and provide a plot and histogram with secondary data (e.g. from fitting plot or histogram).

    This is a specialized analysis class implementing an example analysis routine. For illustrating the analysis
    procedure it only takes a LocData object, creates some random data as result and
    provides plots and a report of the results.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    kwargs : dict
        Parameter that are passed to the algorithm.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : numpy.ndarray, pandas.DataFrame
        Computed results.
    """
    count = 0

    def __init__(self, meta=None, limits=(0, 10)):
        super().__init__(meta=meta, limits=limits)

    def compute(self, locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
          Localization data that might be clustered.

        Returns
        -------
        Analysis class
          Returns the Analysis class object (self).
        """
        data = locdata  # take certain elements from locdata
        self.results = _algorithm_1(data=data, **self.parameter)  # some complicated algorithm
        return self

    def plot(self, ax=None):
        plot(self, ax)

    def plot_2(self, ax=None, bins='auto', normed=True, log=False, fit=True):
        plot_2(self, ax, bins, normed, log, fit)

    def plot_histogram_fit(self, ax=None):
        plot_histogram_fit(self, ax)

    def report(self, path=None):
        report(self, path)


class AnalysisExampleAlgorithm_2(_Analysis):
    """
    Example for an analysis class implementing algorithm_2.
    Compute some data and provide a plot and histogram with secondary data (e.g. from fitting plot or histogram).

    This is a specialized analysis class implementing an example analysis routine. For illustrating the analysis
    procedure it only takes a LocData object, creates some random data as result and
    provides plots and a report of the results.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    kwargs : dict
        Parameter that are passed to the algorithm.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : numpy.ndarray, pandas.DataFrame
        Computed results.
    """
    count = 0

    def __init__(self, meta=None, n_sample=100, seed=None):
        super().__init__(meta=meta, n_sample=n_sample, seed=seed)

    def compute(self, locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
          Localization data that might be clustered.

        Returns
        -------
        Analysis class
          Returns the Analysis class object (self).
        """
        data = locdata  # take certain elements from locdata
        self.results = _algorithm_2(data=data, **self.parameter)  # some complicated algorithm
        return self

    def plot(self, ax=None):
        plot(self, ax)

    def plot_2(self, ax=None, bins='auto', density=True, log=False, fit=True):
        plot_2(self, ax, bins, density, log, fit)

    def plot_histogram_fit(self, ax=None):
        plot_histogram_fit(self, ax)

    def report(self, path=None):
        report(self, path)

#
#### Interface functions
#
# Now we have a class structure with results that can be further processed.
# Secondary results from e.g. fit procedures are added to the analysis class as new attributes.
#


def plot(self, ax=None):
    """
    A specialized plot to give a standardized visualization of results.
    """
    if ax is None:
        ax = plt.gca()

    ax.plot(self.results)
    ax.set(title='Normal Data',
           xlabel=property,
           ylabel='PDF'
           )

    return ax


def plot_2(self, ax=None, bins='auto', density=True, log=False, fit=True):
    """
    A specialized plot to give a standardized visualization of results - in this case a histogram of results.
    """
    if ax is None:
        fig = plt.figure(figsize=(8, 3))
        ax = fig.subplots(nrows=1, ncols=2)
        plt.subplots_adjust(wspace=0)

    # create histogram on first axes
    hist, bins, _ = ax[0].hist(self.results.values, bins=bins, density=density, log=log,
                               label=list(self.results))
    ax[0].set(title='Normal Data',
              xlabel='property',
              ylabel='PDF'
              )

    # create legend and results text on second axes
    h, l = ax[0].get_legend_handles_labels()
    ax[1].legend(h, l,
                 loc='upper left',
                 bbox_to_anchor=(0, 1),
                 title='Legend',
                 frameon=False,
                 borderaxespad=0)

    ax[1].set_axis_off()

    # fit distributions
    if fit:
        plot_histogram_fit(self, ax=ax)

    return ax


def plot_histogram_fit(self, ax=None):
    """
    A specialized plot to give a standardized visualization of results - in this case a histogram of results.
    """
    if ax is None:
        fig = plt.figure(figsize=(8, 3))
        ax = fig.subplots(nrows=1, ncols=2)
        plt.subplots_adjust(wspace=0)

    # fit distributions
    loc, scale = fit_histogram(self, data=self.results['a'].values, id_='a')

    # plot fit
    x_values = np.linspace(stats.norm.ppf(0.01, loc=loc, scale=scale),
                           stats.norm.ppf(0.99, loc=loc, scale=scale), 100)
    ax[0].plot(x_values, stats.norm.pdf(x_values, loc=loc, scale=scale), 'r-', lw=3, alpha=0.6,
               label='norm pdf')

    # present fit results
    ax[1].text(0, 0.5, 'Fit Results:')
    ax[1].text(0, 0.5,
               'center: ' + str(loc) + '\n' + 'sigma: ' + str(scale),
               horizontalalignment='left',
               verticalalignment='top',
               transform=ax[1].transAxes,
               clip_on=False
               )

    ax[1].set_axis_off()

    return ax


def fit_histogram(self, data, id_):

    # MLE fit of distribution on data
    loc, scale = stats.norm.fit(data)

    attribute_center = id_ + '_center'
    attribute_sigma = id_ + '_sigma'
    self.attribute_center = loc
    self.attribute_sigma = scale

    return loc, scale

# there will be other specific visualization methods for other analysis routines.


def report(self, path=None):
    """
    Provide a report that is either displayed or saved as pdf.
    The report is a figure summarizing all visual representations. It is arranged specifically for a particular
    analysis routine.
    """
    fig = plt.figure(figsize=(8.3, 11.7))
    ax = fig.subplots(nrows=3, ncols=2)

    # provide the axes elements (i.e. the plots)
    self.plot(ax=ax[0][0])
    self.plot_2(ax=ax[1][0:2])

    # adjust figure layout
    plt.tight_layout()

    # save figure as pdf
    if path is not None:
        plt.savefig(fname=path, dpi=None, facecolor='w', edgecolor='w',
                    orientation='portrait', papertype=None, format=None,
                    transparent=False, bbox_inches=None, pad_inches=0.1,
                    frameon=None)

    return ax



"""
Alpha shape utility functions for 2d
"""
import numpy as np


__all__ = []


def _circumcircle(points, simplex):
    """
    Center and radius of circumcircle for one triangle.

    Parameters
    -----------
    points : array of shape (n_points, 2)
        point coordinates
    simplex : list
        list with three indices representing a triangle from three points.

    Returns
    -------
    tuple of float
        Center and radius of circumcircle
    """
    A = np.asarray(points)[simplex]
    M = np.array([np.linalg.norm(A, axis=1)**2, A[:, 0], A[:, 1], np.ones(3)], dtype=np.float32)
    S = np.array([0.5 * np.linalg.det(M[[0, 2, 3]]), -0.5 * np.linalg.det(M[[0, 1, 3]])])
    a = np.linalg.det(M[1:])
    b = np.linalg.det(M[[0, 1, 2]])
    return S / a, np.sqrt(b / a + np.linalg.norm(S)**2 / a**2)  # center, radius


def _half_distance(points):
    """
    Half the distance between two points.

    Parameters
    -----------
    points : array of shape (2, 2)
        point coordinates representing a line

    Returns
    -------
    float
        Half the distance between the two points.
    """
    points = np.asarray(points)
    return np.sqrt((points[1, 0] - points[0, 0])**2 + (points[1, 1] - points[0, 1])**2) / 2  # radius = _half_distance
    # this is faster than using: np.linalg.norm(np.diff(A, axis=0)) / 2



import os
from pathlib import Path

import pytest

from locan.dependencies import HAS_DEPENDENCY
from locan import ROOT_DIR
from locan.scripts.script_rois import sc_draw_roi_napari

if HAS_DEPENDENCY["napari"]: import napari


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_rois(capfd):
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs.txt')
    assert path.exists()
    roi_path_list = sc_draw_roi_napari(file_path=str(path), file_type=1, bin_size=20, rescale=None)
    captured = capfd.readouterr()
    print(captured.out)
    assert '[Roi' in captured.out

    # delete roi files
    for path in roi_path_list:
        Path(path).unlink()



"""
Coordinate-based colocalization.

Colocalization is estimated by computing a colocalization index for each localization
using the so-called coordinate-based colocalization algorithm [1]_.

References
----------
.. [1] Malkusch S, Endesfelder U, Mondry J, Gelléri M, Verveer PJ, Heilemann M.,
   Coordinate-based colocalization analysis of single-molecule localization microscopy data.
   Histochem Cell Biol. 2012, 137(1):1-10.
   doi: 10.1007/s00418-011-0880-5
"""
import logging

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors
from scipy.stats import spearmanr
import matplotlib.pyplot as plt

from locan.analysis.analysis_base import _Analysis


__all__ = ['CoordinateBasedColocalization']

logger = logging.getLogger(__name__)


##### The algorithm

def _coordinate_based_colocalization(points, other_points=None, radius=100, n_steps=10):
    """
    Compute a colocalization index for each localization by coordinate-based colocalization.

    Parameters
    ----------
    points : array of tuple
        Array of points (each point represented by a tuple with coordinates) for which CBC values are computed.

    other_points : array of tuple, None
        Array of points (each represented by a tuple with coordinates) to be compared with points. If None other_points
        are set to points.

    radius : int, float
        The maximum radius up to which nearest neighbors are determined

    n_steps : int
        The number of bins from which Spearman correlation is computed.

    Returns
    -------
    np.array
        An array with coordinate-based colocalization coefficients for each input point.
    """
    # sampled radii
    radii = np.linspace(0, radius, n_steps+1)

    # nearest neighbors within radius
    nneigh_1 = NearestNeighbors(radius=radius, metric='euclidean').fit(points)
    distances_1 = np.array(nneigh_1.radius_neighbors()[0])

    if other_points is None:
        nneigh_2 = NearestNeighbors(radius=radius, metric='euclidean').fit(points)
    else:
        nneigh_2 = NearestNeighbors(radius=radius, metric='euclidean').fit(other_points)
    distances_2 = np.array(nneigh_2.radius_neighbors(points)[0])

    # CBC for each point
    correlation = np.empty(len(points))
    for i, (d_1, d_2) in enumerate(zip(distances_1, distances_2)):
        if len(d_1) and len(d_2):
            # binning
            hist_1 = np.histogram(d_1, bins=radii, range=(0, radius))[0]
            hist_2 = np.histogram(d_2, bins=radii, range=(0, radius))[0]

            # normalization
            values_1 = np.cumsum(hist_1) * radius ** 2 / radii[1:] ** 2 / len(d_1)
            values_2 = np.cumsum(hist_2) * radius ** 2 / radii[1:] ** 2 / len(d_2)

            # Spearman rank correlation
            rho, pval = spearmanr(values_1, values_2)
            correlation[i] = rho
        else:
            correlation[i] = np.nan

    # CBC normalization for each point
    max_distances = np.array([np.max(d, initial=0) for d in distances_2])  # max is set to 0 for empty arrays.
    norm_spearmanr = np.exp(-1 * max_distances / radius)
    correlation = correlation * norm_spearmanr

    return correlation


##### The specific analysis classes

class CoordinateBasedColocalization(_Analysis):
    """
    Compute a colocalization index for each localization by coordinate-based colocalization (CBC).

    The colocalization index is calculated for each localization in `locdata` by finding nearest neighbors in
    `locdata` or `other_locdata` within `radius`. A normalized number of nearest neighbors at a certain radius is
    computed for `n_steps` equally-sized steps of increasing radii ranging from 0 to `radius`.
    The Spearman rank correlation coefficent is computed for these values and weighted by
    Exp[-nearestNeighborDistance/distanceMax].

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radius : int, float
        The maximum radius up to which nearest neighbors are determined
    n_steps : int
        The number of bins from which Spearman correlation is computed.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Coordinate-based colocalization coefficients for each input point.
    """
    count = 0

    def __init__(self, meta=None, radius=100, n_steps=10):
        super().__init__(meta=meta, radius=radius, n_steps=n_steps)
        self.results = None

    def compute(self, locdata, other_locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data for which CBC values are computed.
        other_locdata : LocData, None
            Localization data to be colocalized. If None other_locdata is set to locdata.

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
            id_ = other_locdata.meta.identifier
        else:
            other_points = None
            id_ = 'self'

        self.results = pd.DataFrame({f'colocalization_cbc_{id_}':
                                         _coordinate_based_colocalization(points, other_points, **self.parameter)})
        return self

    def hist(self, ax=None, bins=(-1, -0.3333, 0.3333, 1), density=True, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing hist(results).

        Parameters
        ----------
        bins : int, list, 'auto'
            Bin specification as used in :func:`matplotlib.hist`.
        density : bool
            Flag for normalization as used in :func:`matplotlib.hist`. True returns probability density function; None returns
            counts.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        ax.hist(self.results.iloc[:, 0].values, bins=bins, density=density, label='cbc')

        ax.set(title='CBC histogram',
               xlabel='colocalization_cbc',
               ylabel='pdf' if density else 'counts'
               )
        ax.legend(loc='best')

        return ax



"""
Base class that serves as template for a specialized analysis class.

It also provides helper functions to be used in specialized analysis classes.
"""
import time

from google.protobuf import text_format, json_format
from scipy import stats

from locan.analysis import metadata_analysis_pb2
from locan.utils.format import _time_string


class _Analysis:
    """
    Base class for standard analysis procedures.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    kwargs :
        Parameter that are passed to the algorithm.

    Attributes
    ----------
    locdata : LocData
        Localization data.
    parameter : dict
        A dictionary with all settings (i.e. the kwargs) for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : any
        Computed results.
    """
    count = 0
    """int: A counter for counting Analysis class instantiations (class attribute)."""

    def __init__(self, meta, **kwargs):
        self.__class__.count += 1

        self.parameter = kwargs
        self.meta = _init_meta(self)
        self.meta = _update_meta(self, meta)
        self.results = None

    def __del__(self):
        """ Update the counter upon deletion of class instance. """
        self.__class__.count -= 1

    def __repr__(self):
        """ Return representation of the Analysis class. """
        parameter_string = ", ".join((f'{key}={value}' for key, value in self.parameter.items()))
        return f'{self.__class__.__name__}({parameter_string})'

    def __getstate__(self):
        """Modify pickling behavior."""
        # Copy the object's state from self.__dict__ to avoid modifying the original state.
        state = self.__dict__.copy()
        # Serialize the unpicklable protobuf entries.
        json_string = json_format.MessageToJson(self.meta, including_default_value_fields=False)
        state['meta'] = json_string
        return state

    def __setstate__(self, state):
        """Modify pickling behavior."""
        # Restore instance attributes.
        self.__dict__.update(state)
        # Restore protobuf class for meta attribute
        self.meta = metadata_analysis_pb2.AMetadata()
        self.meta = json_format.Parse(state['meta'], self.meta)

    def __bool__(self):
        if self.results is not None:
            return True
        else:
            return False

    def compute(self):
        """ Apply analysis routine with the specified parameters on locdata and return results."""
        raise NotImplementedError

    def report(self):
        """ Show a report about analysis results."""
        raise NotImplementedError


# Dealing with metadata

def _init_meta(self):
    meta_ = metadata_analysis_pb2.AMetadata()
    meta_.identifier = str(self.__class__.count)
    meta_.creation_time.GetCurrentTime()
    meta_.method.name = str(self.__class__.__name__)
    meta_.method.parameter = str(self.parameter)
    return meta_


def _update_meta(self, meta=None):
    meta_ = self.meta
    if meta is None:
        pass
    else:
        try:
            meta_.MergeFrom(meta)
        except TypeError:
            for key, value in meta.items():
                setattr(meta_, key, value)

    return meta_


# Dealing with scipy.stats

def _list_parameters(distribution):
    """
    List parameters for scipy.stats.distribution.

    Parameters
    ----------
    distribution : str, scipy.stats.distribution
        Distribution of choice.

    Returns
    -------
    list of str
        A list of distribution parameter strings.
    """
    if isinstance(distribution, str):
        distribution = getattr(stats, distribution)
    if distribution.shapes:
        parameters = [name.strip() for name in distribution.shapes.split(',')]
    else:
        parameters = []
    if distribution.name in stats._discrete_distns._distn_names:
        parameters += ['loc']
    elif distribution.name in stats._continuous_distns._distn_names:
        parameters += ['loc', 'scale']
    else:
        raise TypeError("Distribution name not found in discrete or continuous lists.")

    return parameters



from copy import deepcopy

import pytest
import numpy as np

from locan.dependencies import HAS_DEPENDENCY
from locan.data.cluster.clustering import cluster_hdbscan, cluster_dbscan
from locan.data.cluster.utils import serial_clustering


# tests hdbscan
@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_2d(locdata_two_cluster_2d):
    noise, clust = cluster_hdbscan(locdata_two_cluster_2d, min_cluster_size=2, allow_single_cluster=False)
    assert noise is None
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_hdbscan(locdata_two_cluster_2d, min_cluster_size=5, allow_single_cluster=True)
    assert (len(clust) == 1)
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_hdbscan(locdata_two_cluster_2d, loc_properties=['position_x'],
                                   min_cluster_size=2, allow_single_cluster=False)
    assert noise is None
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_2d_with_noise(locdata_two_cluster_with_noise_2d):
    noise, clust = cluster_hdbscan(locdata_two_cluster_with_noise_2d, min_cluster_size=2, allow_single_cluster=False)
    assert len(noise) == 1
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_hdbscan(locdata_two_cluster_with_noise_2d, min_cluster_size=5, allow_single_cluster=True)
    assert len(noise) == 1
    assert (len(clust) == 1)
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_with_shuffled_index(locdata_two_cluster_with_noise_2d):
    locdata = deepcopy(locdata_two_cluster_with_noise_2d)
    new_index = list(locdata.data.index)
    np.random.shuffle(new_index)
    locdata.data.index = new_index
    noise, clust = cluster_hdbscan(locdata, min_cluster_size=2, allow_single_cluster=False)
    assert len(noise) == 1
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', (0, 0)),
    ('locdata_single_localization', (1, 0))
])
def test_cluster_hdbscan_empty_locdata(
        locdata_empty, locdata_single_localization,
        fixture_name, expected):
    locdata = eval(fixture_name)
    noise, clust = cluster_hdbscan(locdata, min_cluster_size=2, allow_single_cluster=False)
    assert len(noise) == expected[0]
    assert len(clust) == expected[1]


# tests dbscan

def test_cluster_dbscan_2d(locdata_two_cluster_2d):
    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=2, min_samples=1)
    assert noise is None
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=20, min_samples=1)
    assert (len(clust) == 1)
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_2d, loc_properties=['position_x'], eps=2, min_samples=1)
    assert noise is None
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=0.2, min_samples=10)
    assert len(noise) == len(locdata_two_cluster_2d)
    assert len(clust) == 0
    assert noise.region == locdata_two_cluster_2d.region
    assert clust.region is None


def test_cluster_dbscan_2d_with_noise(locdata_two_cluster_with_noise_2d):
    noise, clust = cluster_dbscan(locdata_two_cluster_with_noise_2d, eps=2, min_samples=1)
    assert noise is None
    assert (len(clust) == 3)
    assert all(clust.data.localization_count == [3, 3, 1])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_with_noise_2d, eps=2, min_samples=2)
    assert len(noise) == 1
    assert (len(clust) == 2)
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_with_noise_2d, eps=20, min_samples=2)
    assert len(noise) == 1
    assert (len(clust) == 1)
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


def test_cluster_dbscan_with_shuffled_index(locdata_two_cluster_with_noise_2d):
    locdata = deepcopy(locdata_two_cluster_with_noise_2d)
    new_index = list(locdata.data.index)
    np.random.shuffle(new_index)
    locdata.data.index = new_index
    noise, clust = cluster_dbscan(locdata, eps=2, min_samples=2)
    assert len(noise) == 1
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)#
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', (0, 0)),
    ('locdata_single_localization', (1, 0))
])
def test_cluster_dbscan_empty_locdata(
        locdata_empty, locdata_single_localization,
        fixture_name, expected):
    locdata = eval(fixture_name)
    noise, clust = cluster_dbscan(locdata, eps=2, min_samples=2)
    assert len(noise) == expected[0]
    assert len(clust) == expected[1]


# tests serial_clustering

def test_serial_clustering(locdata_two_cluster_with_noise_2d):
    noise, clust = serial_clustering(locdata_two_cluster_with_noise_2d, cluster_dbscan,
                                     parameter_lists=dict(eps=[2, 20], min_samples=[2, 5]))
    assert (len(noise) == 4)
    assert (len(clust) == 4)



"""

Compute Ripley's k function.

Spatial clustering of localization data is characterized by Ripley's k or related l and h functions [1]_.


Ripley's k function is computed for 2D and 3D data for a series of radii as described in [2]_ in order to provide
evidence for deviations from a spatially homogeneous Poisson process (i.e. complete spatial randomness, CSR).
Ripley' s k function is estimated by summing all points or over test points being a random subset of all points:

.. math::

   k(r) = \\frac{1}{\\lambda (n-1)} \\sum_{i=1}^{n} N_{p_{i}}(r)


here :math:`p_i` is the :math:`i^{th}` point of n test points, :math:`N_{p_{i}}` is the number of points within the region
of radius r around :math:`p_{i}`, and :math:`\\lambda` is the density of all points.

We follow the definition of l and h functions in [2]_. Ripley's l function is:

.. math::

   l(r) &= \\sqrt{k(r)) / \pi} \\qquad \\text{in 2D}

   l(r) &= \\sqrt[3]{\\frac{3}{4 \\pi} k(r)} \\qquad \\text{in 3D}


And Ripley's h function is:

.. math::

   h(r) = l(r) - r


References
----------
.. [1] B.D. Ripley, Modelling spatial patterns. Journal of the Royal Statistical Society, 1977, 172–212.
.. [2] Kiskowski, M. A., Hancock, J. F., and Kenworthy, A. K., On the use of Ripley's K-function and its
   derivatives to analyze domain size. Biophysical journal, 2009, 97, 1095–1103.

"""
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.neighbors import NearestNeighbors

from locan.analysis.analysis_base import _Analysis


__all__ = ['RipleysKFunction', 'RipleysLFunction', 'RipleysHFunction']

logger = logging.getLogger(__name__)

#### The algorithms


def _ripleys_k_function(points, radii, region_measure=1, other_points=None):
    """
    Compute Ripley's K function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    points : array
        2D or 3D points on which to estimate Ripley's K function.
    radii: array of float
        Radii at which to compute Ripley's k function.
    region_measure : float
        region measure (area or volume) for points.
    other_points : array or None
        2D or 3D points from which to estimate Ripley's K function (e.g. subset of points). For None other_points
        is set to points (default).

    Returns
    -------
    ripley : numpy.ndarray of floats with shape (n_radii,)
        Ripley's K function evaluated at input radii.
    """
    if other_points is None:
        other_points = points

    nn = NearestNeighbors(metric='euclidean').fit(points)

    ripley_0 = []
    for radius in radii:
        indices_list = nn.radius_neighbors(other_points, radius, return_distance=False)
        ripley_0.append(np.array([len(indices)-1 for indices in indices_list]).sum())

    ripley = region_measure / (len(points) * (len(other_points) - 1)) * np.array(ripley_0)

    return ripley


def _ripleys_l_function(points, radii, region_measure=1, other_points=None):
    """
    Evaluates Ripley's L function which is different for 2D and 3D data points. For parameter description
    see _ripleys_k_function.
    """
    if np.shape(points)[1] == 2:
        return np.sqrt(_ripleys_k_function(points, radii, region_measure, other_points) / np.pi)
    elif np.shape(points)[1] == 3:
        return np.cbrt(_ripleys_k_function(points, radii, region_measure, other_points) * 3 / 4 / np.pi)


def _ripleys_h_function(points, radii, region_measure=1, other_points=None):
    """
    Evaluates Ripley's H function. For parameter description
    see _ripleys_k_function.
    """

    return _ripleys_l_function(points, radii, region_measure, other_points) - radii


##### The specific analysis classes

class RipleysKFunction(_Analysis):
    """
    Compute Ripley's K function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radii: array of float
        Radii at which to compute Ripley's k function.
    region_measure : float or 'bb'
        Region measure (area or volume) for point region. For 'bb' the region measure of the bounding_box is used.


    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with radii as provided and Ripley's K function.
    """
    count = 0

    def __init__(self, meta=None, radii=np.linspace(0, 100, 10), region_measure='bb'):
        super().__init__(meta=meta, radii=radii, region_measure=region_measure)
        self.results = None

    def compute(self, locdata, other_locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data with 2D or 3D coordinates on which to estimate Ripley's K function.
        other_locdata : LocData
            Other localization data from which to estimate Ripley's K function (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter['region_measure'] == 'bb':
            region_measure = float(locdata.properties['region_measure_bb'])
        else:
            region_measure = self.parameter['region_measure']

        ripley = _ripleys_k_function(points=points, radii=self.parameter['radii'],
                                           region_measure=region_measure, other_points=other_points)
        self.results = pd.DataFrame({'radius': self.parameter['radii'], 'Ripley_k_data': ripley})
        self.results = self.results.set_index('radius')
        return self

    def plot(self, ax=None, **kwargs):
        plot(self, ax, **kwargs)


class RipleysLFunction(_Analysis):
    """
    Compute Ripley's L function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radii: array of float
        Radii at which to compute Ripley's k function.
    region_measure : float or 'bb'
        Region measure (area or volume) for point region. For 'bb' the region measure of the bounding_box is used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with radii as provided and Ripley's L function.
    """
    count = 0

    def __init__(self, meta=None, radii=np.linspace(0, 100, 10), region_measure='bb'):
        super().__init__(meta=meta, radii=radii, region_measure=region_measure)
        self.results = None

    def compute(self, locdata, other_locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data with 2D or 3D coordinates on which to estimate Ripley's L function.
        other_locdata : LocData
            Other localization data from which to estimate Ripley's L function (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter['region_measure'] == 'bb':
            region_measure = float(locdata.properties['region_measure_bb'])
        else:
            region_measure = self.parameter['region_measure']

        ripley = _ripleys_l_function(points=points, radii=self.parameter['radii'],
                                           region_measure=region_measure, other_points=other_points)
        self.results = pd.DataFrame({'radius': self.parameter['radii'], 'Ripley_l_data': ripley})
        self.results = self.results.set_index('radius')
        return self

    def plot(self, ax=None, **kwargs):
        plot(self, ax, **kwargs)


class RipleysHFunction(_Analysis):
    """
    Compute Ripley's H function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radii: array of float
        Radii at which to compute Ripley's k function.
    region_measure : float or 'bb'
        Region measure (area or volume) for point region. For 'bb' the region measure of the bounding_box is used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with radii as provided and Ripley's H function.
    Ripley_h_maximum : pandas.DataFrame
        Data frame with radius and Ripley's H value for the radius at which the H function has its maximum.
    """
    count = 0

    def __init__(self, meta=None, radii=np.linspace(0, 100, 10), region_measure='bb'):
        super().__init__(meta=meta, radii=radii, region_measure=region_measure)
        self.results = None
        self._Ripley_h_maximum = None

    def compute(self, locdata, other_locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data with 2D or 3D coordinates on which to estimate Ripley's H function.
        other_locdata : LocData
            Other localization data from which to estimate Ripley's H function (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        # reset secondary results
        self._Ripley_h_maximum = None

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter['region_measure'] == 'bb':
            region_measure = float(locdata.properties['region_measure_bb'])
        else:
            region_measure = self.parameter['region_measure']

        ripley = _ripleys_h_function(points=points, radii=self.parameter['radii'],
                                           region_measure=region_measure, other_points=other_points)
        self.results = pd.DataFrame({'radius': self.parameter['radii'], 'Ripley_h_data': ripley})
        self.results = self.results.set_index('radius')
        return self


    @property
    def Ripley_h_maximum(self):
        if self._Ripley_h_maximum is None:
            index = self.results['Ripley_h_data'].idxmax()
            self._Ripley_h_maximum = pd.DataFrame({'radius': index, 'Ripley_h_maximum': self.results.loc[index]})
            return self._Ripley_h_maximum
        else:
            return self._Ripley_h_maximum

    @Ripley_h_maximum.deleter
    def Ripley_h_maximum(self):
        self._Ripley_h_maximum = None

    def plot(self, ax=None, **kwargs):
        plot(self, ax, **kwargs)


##### Interface functions


def plot(self, ax=None, **kwargs):
    """
    Provide plot of results as :class:`matplotlib.axes.Axes` object.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        The axes on which to show the image
    kwargs : dict
        Other parameters passed to :func:`matplotlib.pyplot.plot`.

    Returns
    -------
    :class:`matplotlib.axes.Axes`
        Axes object with the plot.
    """
    if ax is None:
        ax = plt.gca()

    if not self:
        return ax

    self.results.plot(ax=ax, **kwargs)

    if self.results.columns[0] == 'Ripley_k_data':
        ylabel = 'K-function'
    elif self.results.columns[0] == 'Ripley_l_data':
        ylabel = 'L-function'
    elif self.results.columns[0] == 'Ripley_h_data':
        ylabel = 'H-function'
    else:
        ylabel = None

    ax.set(title="Ripley's " + ylabel,
           xlabel='radius',
           ylabel=ylabel
           )

    return ax



import pytest
import numpy as np

from locan.dependencies import HAS_DEPENDENCY
from locan.data.transform import transform_affine
from locan.data.register import register_icp, register_cc


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_register_icp_2d(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 100))
    offset_target = np.array([100., 100.])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_icp(locdata_blobs_2d, locdata_2d_transformed, verbose=False)
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(locdata_blobs_2d, locdata_2d_transformed, pre_translation=(-90, -90), verbose=False)
    offset_target = np.array([10., 10.])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_register_icp_3d(locdata_blobs_3d):
    locdata_3d_transformed = transform_affine(locdata_blobs_3d, offset=(100, 100, 100))
    offset_target = np.array([100., 100., 100.])
    matrix_target = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    matrix, offset = register_icp(locdata_blobs_3d, locdata_3d_transformed, verbose=False)
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(locdata_blobs_3d, locdata_3d_transformed, pre_translation=(-90, -90, -90),
                                  verbose=False)
    offset_target = np.array([10., 10., 10.])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

def test_register_cc(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 50))
    offset_target = np.array([100., 50.])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_cc(locdata_blobs_2d, locdata_2d_transformed, bin_size=50, verbose=False)
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_cc(locdata_blobs_2d, locdata_2d_transformed, bin_size=(10, 50), verbose=False)
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)



import pytest
import numpy as np
import pandas as pd
from scipy import stats

from locan import LocData
from locan.analysis.blinking import _blink_statistics, _DistributionFits
from locan.analysis import BlinkStatistics


def test__blink_statistics_0():
    # frame with on and off periods up to three frames and starting with one-frame on-period.
    frames = np.array([0, 4, 6, 7, 8, 12, 13])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [1, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [3, 1, 3])
    assert np.array_equal(results['on_periods_frame'], [0, 4, 6, 12])
    assert np.array_equal(results['off_periods_frame'], [1, 5, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0], [1], [2, 3, 4], [5, 6]]
                                                             )])

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [1, 5, 2])
    assert np.array_equal(results['off_periods'], [3, 3])
    assert np.array_equal(results['on_periods_frame'], [0, 4, 12])
    assert np.array_equal(results['off_periods_frame'], [1, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0], [1, 2, 3, 4], [5, 6]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [14])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5, 6]]
                                                             )])


def test__blink_statistics_1():
    # frame with on and off periods up to three frames and starting with two-frame on-period.
    frames = np.array([0, 1, 3, 6, 7, 8, 12, 13])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [1, 2, 3])
    assert np.array_equal(results['on_periods_frame'], [0, 3, 6, 12])
    assert np.array_equal(results['off_periods_frame'], [2, 4, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [4, 3, 2])
    assert np.array_equal(results['off_periods'], [2, 3])
    assert np.array_equal(results['on_periods_frame'], [0, 6, 12])
    assert np.array_equal(results['off_periods_frame'], [4, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [14])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5, 6, 7]]
                                                             )])


def test__blink_statistics_2():
    # frame with on and off periods up to three frames and starting with two-frame on-period.
    frames = np.array([0, 1, 3, 6, 7, 8, 12, 13]) + 1

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [1, 1, 2, 3])
    assert np.array_equal(results['on_periods_frame'], [1, 4, 7, 13])
    assert np.array_equal(results['off_periods_frame'], [0, 3, 5, 10])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [5, 3, 2])
    assert np.array_equal(results['off_periods'], [2, 3])
    assert np.array_equal(results['on_periods_frame'], [0, 7, 13])
    assert np.array_equal(results['off_periods_frame'], [5, 10])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [15])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5, 6, 7]]
                                                             )])


def test__blink_statistics_3():
    # frame with on and off periods up to three frames and starting with off-period.
    frames = np.array([0, 1, 4, 6, 7, 8, 12, 13]) + 4

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [4, 2, 1, 3])
    assert np.array_equal(results['on_periods_frame'], [4, 8, 10, 16])
    assert np.array_equal(results['off_periods_frame'], [0, 6, 9, 13])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=True)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [2, 1, 3])
    assert np.array_equal(results['on_periods_frame'], [4, 8, 10, 16])
    assert np.array_equal(results['off_periods_frame'], [6, 9, 13])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=2, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [9, 2])
    assert np.array_equal(results['off_periods'], [4, 3])
    assert np.array_equal(results['on_periods_frame'], [4, 16])
    assert np.array_equal(results['off_periods_frame'], [0, 13])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=2, remove_heading_off_periods=True)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [9, 2])
    assert np.array_equal(results['off_periods'], [3])
    assert np.array_equal(results['on_periods_frame'], [4, 16])
    assert np.array_equal(results['off_periods_frame'], [13])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5], [6, 7]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [18])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert np.array_equal(results['on_periods_indices'], [[0, 1, 2, 3, 4, 5, 6, 7]])


def test__blink_statistics_4():
    # frame with on and off periods up to three frames and starting with off-period.
    frames = np.array([0, 1, 4, 6, 12, 13]) + 2

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 1, 2])
    assert np.array_equal(results['off_periods'], [2, 2, 1, 5])
    assert np.array_equal(results['on_periods_frame'], [2, 6, 8, 14])
    assert np.array_equal(results['off_periods_frame'], [0, 4, 7, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3], [4, 5]]
                                                             )])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=True)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 1, 2])
    assert np.array_equal(results['off_periods'], [2, 1, 5])
    assert np.array_equal(results['on_periods_frame'], [2, 6, 8, 14])
    assert np.array_equal(results['off_periods_frame'], [4, 7, 9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3], [4, 5]]
                                                             )])

    results = _blink_statistics(frames, memory=3, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [9, 2])
    assert np.array_equal(results['off_periods'], [5])
    assert np.array_equal(results['on_periods_frame'], [0, 14])
    assert np.array_equal(results['off_periods_frame'], [9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3], [4, 5]]
                                                             )])

    results = _blink_statistics(frames, memory=3, remove_heading_off_periods=True)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [7, 2])
    assert np.array_equal(results['off_periods'], [5])
    assert np.array_equal(results['on_periods_frame'], [2, 14])
    assert np.array_equal(results['off_periods_frame'], [9])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3], [4, 5]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [16])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5]]
                                                             )])


def test__blink_statistics_5(caplog):
    # frame with on and off periods including repeated frames.
    frames = np.array([0, 1, 4, 4, 6, 7, 8, 12, 12, 13]) + 4

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [2, 1, 3, 2])
    assert np.array_equal(results['off_periods'], [4, 2, 1, 3])
    assert np.array_equal(results['on_periods_frame'], [4, 8, 10, 16])
    assert np.array_equal(results['off_periods_frame'], [0, 6, 9, 13])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1], [2], [3, 4, 5], [6, 7]]
                                                             )])
    assert caplog.record_tuples == [('locan.analysis.blinking', 30,
                                     'There are 2 duplicated frames found that will be ignored.')]

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [18])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3, 4, 5, 6, 7]]
                                                             )])


def test__blink_statistics_6():
    # frame with on and off periods up to three frames and starting with one-frame on-period.
    frames = np.array([0, 2, 3, 9])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [1, 2, 1])
    assert np.array_equal(results['off_periods'], [1, 5])
    assert np.array_equal(results['on_periods_frame'], [0, 2, 9])
    assert np.array_equal(results['off_periods_frame'], [1, 4])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0], [1, 2], [3]]
                                                             )])

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results['on_periods']) == len(results['on_periods_frame'])
    assert len(results['off_periods']) == len(results['off_periods_frame'])
    assert np.array_equal(results['on_periods'], [10])
    assert np.array_equal(results['off_periods'], [])
    assert np.array_equal(results['on_periods_frame'], [0])
    assert np.array_equal(results['off_periods_frame'], [])
    assert all([np.array_equal(one, two) for one, two in zip(results['on_periods_indices'],
                                                             [[0, 1, 2, 3]]
                                                             )])


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1],
        'intensity': [0, 1, 3, 4, 1],
        'psf_sigma_x': [100, 100, 100, 100, 100],
        }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_with_zero_frame():
    locdata_dict = {
        'frame': [0, 1, 2, 4, 10, 11, 14]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_without_zero_frame():
    locdata_dict = {
        'frame': [1, 2, 4, 10, 11, 14]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_with_repetitions():
    locdata_dict = {
        'frame': [2, 2, 2, 4, 4, 14]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_blink_statistics(locdata_with_zero_frame, locdata_without_zero_frame):
    bs = _blink_statistics(locdata_with_zero_frame, memory=0, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [3, 1, 2, 1])
    assert all(bs['off_periods'] == [1, 5, 2])

    bs = _blink_statistics(locdata_with_zero_frame.data.frame.values, memory=0, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [3, 1, 2, 1])
    assert all(bs['off_periods'] == [1, 5, 2])

    bs = _blink_statistics(locdata_without_zero_frame, memory=0, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [2, 1, 2, 1])
    assert all(bs['off_periods'] == [1, 1, 5, 2])

    bs = _blink_statistics(locdata_with_zero_frame, memory=0, remove_heading_off_periods=True)
    assert all(bs['on_periods'] == [3, 1, 2, 1])
    assert all(bs['off_periods'] == [1, 5, 2])

    bs = _blink_statistics(locdata_without_zero_frame, memory=0, remove_heading_off_periods=True)
    assert all(bs['on_periods'] == [2, 1, 2, 1])
    assert all(bs['off_periods'] == [1, 5, 2])

    bs = _blink_statistics(locdata_with_zero_frame, memory=1, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [5, 2, 1])
    assert all(bs['off_periods'] == [5, 2])

    bs = _blink_statistics(locdata_without_zero_frame, memory=1, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [5, 2, 1])
    assert all(bs['off_periods'] == [5, 2])

    bs = _blink_statistics(locdata_with_zero_frame, memory=2, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [5, 5])
    assert all(bs['off_periods'] == [5])

    bs = _blink_statistics(locdata_without_zero_frame, memory=2, remove_heading_off_periods=False)
    assert all(bs['on_periods'] == [5, 5])
    assert all(bs['off_periods'] == [5])


def test_blink_statistics__with_repetitions(locdata_with_repetitions):
    _blink_statistics(locdata_with_repetitions, memory=0, remove_heading_off_periods=False)


def test_BlinkStatistics_empty(caplog):
    bs = BlinkStatistics().compute(LocData())
    bs.fit_distributions()
    bs.hist()
    assert caplog.record_tuples == [('locan.analysis.blinking', 30, 'Locdata is empty.'),
                                    ('locan.analysis.blinking', 30, 'No results available to fit.')]


def test_BlinkStatistics(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    assert repr(bs) == "BlinkStatistics(memory=0, remove_heading_off_periods=True)"
    assert all(bs.results['on_periods'] == [3, 1, 2, 1])
    assert all(bs.results['off_periods'] == [1, 5, 2])
    assert bs.distribution_statistics == {}

    bs.hist(data_identifier='on_periods', ax=None, bins='auto', log=True, fit=False)
    bs.hist(data_identifier='off_periods', ax=None, bins='auto', log=True, fit=False)
    bs.hist(data_identifier='on_periods', ax=None, bins='auto', log=True, fit=True)


def test_DistributionFits(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    df = _DistributionFits(bs, distribution=stats.expon, data_identifier='on_periods')
    # print(df.analysis_class.results)
    assert len(df.analysis_class.results) == 5
    assert df.data_identifier == 'on_periods'
    assert repr(df) == "_DistributionFits(analysis_class=BlinkStatistics, " \
                       "distribution=expon_gen, data_identifier=on_periods)"
    assert df.parameter_dict() == {}
    df.fit()
    assert list(df.parameter_dict().keys()) == ['on_periods_loc', 'on_periods_scale']

    df = _DistributionFits(bs, distribution=stats.expon, data_identifier='off_periods')
    df.fit()
    assert list(df.parameter_dict().keys()) == ['off_periods_loc', 'off_periods_scale']
    df.plot()

    # print(df.analysis_class.results[df.data_identifier])


def test_fit_distributions(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions()
    assert bs.distribution_statistics['on_periods'].parameter_dict() == \
           {'on_periods_loc': 1.0, 'on_periods_scale': 0.75}
    assert bs.distribution_statistics['off_periods'].parameter_dict() == \
           {'off_periods_loc': 1.0, 'off_periods_scale': 1.6666666666666665}
    bs.hist()
    bs.hist(data_identifier='off_periods')
    del bs

    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions(with_constraints=False)
    assert bs.distribution_statistics['on_periods'].parameter_dict()['on_periods_loc'] == 1
    assert bs.distribution_statistics['off_periods'].parameter_dict()['off_periods_loc'] == 1
    del bs

    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions(data_identifier='on_periods')
    assert bs.distribution_statistics['on_periods'].parameter_dict() == \
           {'on_periods_loc': 1.0, 'on_periods_scale': 0.75}



"""

Analyze localization property.

Localizations come with a range of properties including position coordinates, emission strength, local background etc..
Most properties represent random variables that were drawn from an unknown probability distribution.
It is often useful to analyze the properties from all localizations within a selection and estimate the corresponding
probability distribution.

"""
import warnings
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.neighbors import NearestNeighbors
from scipy import stats

from locan.analysis.analysis_base import _Analysis, _list_parameters


__all__ = ['LocalizationProperty']

logger = logging.getLogger(__name__)


##### The algorithms

def _localization_property(locdata, loc_property='intensity', index=None):
    if index is None:
        results = locdata.data[[loc_property]]
    else:
        results = locdata.data[[loc_property, index]].set_index(index)
        results = results.sort_index()

    return results


##### The specific analysis classes

class LocalizationProperty(_Analysis):
    """
    Analyze localization property with respect to probability density or variation over a specified index.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_property : str
        The property to analyze.
    index : str, None
        The property name that should serve as index (i.e. x-axis in x-y-plot)

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : numpy.ndarray, pandas.DataFrame
        Computed results.
    distribution_statistics : Distribution_stats, None
        Distribution parameters derived from MLE fitting of results.
    """
    def __init__(self, meta=None, loc_property='intensity', index=None):
        super().__init__(meta=meta, loc_property=loc_property, index=index)
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _localization_property(locdata=locdata, **self.parameter)
        return self

    def fit_distributions(self, distribution=stats.expon, with_constraints=True, **kwargs):
        """
        Fit probability density functions to the distributions of `loc_property` values in the results
        using MLE (scipy.stats).

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then `floc=np.min(self.analysis_class.results[self.loc_property].values)`.

        Parameters
        ----------
        distribution : str, scipy.stats distribution
            Distribution model to fit.
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        kwargs : dict
            Other parameters are passed to the `scipy.stat.distribution.fit()` function.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(distribution, with_constraints=with_constraints, **kwargs)
        else:
            logger.warning('No results available to fit.')

    def plot(self, ax=None, window=1, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the running average of results over window size.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        window: int
            Window for running average that is applied before plotting.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        self.results.rolling(window=window, center=True).mean().plot(ax=ax, **dict(dict(legend=False), **kwargs))
        ax.set(title=f"{self.parameter['loc_property']}({self.parameter['index']})\n (window={window})",
               xlabel=self.parameter['index'],
               ylabel=self.parameter['loc_property']
               )

        return ax

    def hist(self, ax=None, bins='auto', log=True, fit=True, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing hist(results). Nan entries are ignored.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        bins : float
            Bin specifications (passed to :func:`matplotlib.hist`).
        log : Bool
            Flag for plotting on a log scale.
        fit: bool, None
            Flag indicating if distribution fit is shown. The fit will only be computed if `distribution_statistics`
             is None.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        ax.hist(self.results.dropna(axis=0).values, bins=bins, **dict(dict(density=True, log=log), **kwargs))
        ax.set(title=self.parameter['loc_property'],
               xlabel=self.parameter['loc_property'],
               ylabel='PDF'
               )

        # fit distributions:
        if fit:
            if isinstance(self.distribution_statistics, _DistributionFits):
                self.distribution_statistics.plot(ax=ax)
            else:
                self.fit_distributions()
                self.distribution_statistics.plot(ax=ax)

        return ax


# todo add Dependence_stats to fit a plot to a linear function, log function, or exponential decay.

class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by LocalizationProperty methods.
    It holds the statistical parameters derived by fitting the result distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationPrecision object
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.
    loc_property : str
        The LocData property for which to fit an appropriate distribution.
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    parameters : list of str
        Distribution parameters.
    """
    def __init__(self, analysis_class):
        self.analysis_class = analysis_class
        self.loc_property = self.analysis_class.parameter['loc_property']
        self.distribution = None
        self.parameters = []

    def fit(self, distribution, with_constraints=True, **kwargs):
        """
        Fit scipy.stats.distribution to analysis_class.results[loc_property].

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        kwargs : dict
            Other parameters are passed to the `scipy.stat.distribution.fit()` function.
        """
        self.distribution = distribution
        for param in _list_parameters(distribution):
            self.parameters.append(self.loc_property + '_' + param)

        if with_constraints and self.distribution == stats.expon:
            # MLE fit of exponential distribution with constraints
            fit_results = stats.expon.fit(self.analysis_class.results[self.loc_property].values,
                                          **dict(dict(floc=np.min(self.analysis_class.results[self.loc_property].values)
                                                      ), **kwargs))
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)
        else:
            fit_results = self.distribution.fit(self.analysis_class.results[self.loc_property].values, **kwargs)
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)

    def plot(self, ax=None, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the probability distribution functions of fitted
        results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        parameter = self.parameter_dict().values()
        x_values = np.linspace(self.distribution.ppf(0.001, *parameter),
                               self.distribution.ppf(0.999, *parameter), 100)
        ax.plot(x_values, self.distribution.pdf(x_values, *parameter), 'r-',
                **dict(dict(lw=3, alpha=0.6, label=str(self.distribution.name) + ' pdf'), **kwargs))

        return ax

    def parameter_dict(self):
        """ Dictionary of fitted parameters. """
        return {k: self.__dict__[k] for k in self.parameters}



import pytest
import numpy as np
import pandas as pd

from locan import LocData
from locan.dependencies import HAS_DEPENDENCY
if HAS_DEPENDENCY["trackpy"]:
    from locan.data.tracking import link_locdata, track
    from trackpy import quiet as tp_quiet


pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["trackpy"], reason="requires trackpy")

if HAS_DEPENDENCY["trackpy"]:
    tp_quiet()  # same as: trackpy.logger.setLevel(logging.WARN)


@pytest.fixture()
def locdata_simple():
    dict_ = {
        'position_x': [0, 1, 2, 10, 20, 21, 30, 4],
        'position_y': [0, 1, 2, 10, 20, 21, 30, 4],
        'position_z': [0, 1, 2, 10, 20, 21, 30, 4],
        'frame': np.arange(8),

    }
    return LocData(dataframe=pd.DataFrame.from_dict(dict_))


def test_link_locdata(locdata_simple):
    track_series = link_locdata(locdata_simple, search_range=5, memory=0)
    assert (len(track_series) == 8)
    assert(track_series.name == 'track')


def test_track(locdata_simple):
    locdata_new, track_series = track(locdata_simple, search_range=5)
    # print(locdata_new.data)
    assert 'frame' in locdata_new.data.columns
    assert (len(locdata_new) == 5)
    locdata_new, track_series = track(locdata_simple, search_range=5, memory=5)
    # print(locdata_new.data)
    assert (len(locdata_new) == 4)



import pytest
from locan.scripts import script_show_versions


def test_script_show_version(capfd):
    script_show_versions.main([])
    captured = capfd.readouterr()
    for header in ['\nLocan:', '\nPython:', '\nSystem:', '\nPython dependencies:']:
        assert header in captured.out
    for header in ['node:', 'executable:', 'locan']:
        assert header not in captured.out

    script_show_versions.main(["-o", "locan"])
    captured = capfd.readouterr()
    for header in ['\nLocan:', '\nPython:', '\nSystem:', '\nPython dependencies:', 'locan']:
        assert header in captured.out

    script_show_versions.main(["-v", "-e", "-o", "locan"])
    captured = capfd.readouterr()
    for header in ['\nLocan:', '\nPython:', '\nSystem:', '\nPython dependencies:',
                   'locan', 'node:', 'executable:']:
        assert header in captured.out



from pathlib import Path
import tempfile
import pickle

from locan.analysis.analysis_example import AnalysisExampleAlgorithm_1, AnalysisExampleAlgorithm_2
from locan.analysis import metadata_analysis_pb2


def test_Analysis_example():
    ae = AnalysisExampleAlgorithm_1(limits=(100, 110), meta={'comment': 'this is an example'})
    # print(ae)
    assert repr(ae) == 'AnalysisExampleAlgorithm_1(limits=(100, 110))'
    assert str(ae) == repr(ae)
    assert ae.parameter == {'limits': (100, 110)}
    assert ae.meta.comment == 'this is an example'
    ae.compute(locdata=None)
    assert (len(ae.results) == 10)

    ae.plot()
    ae.plot_2()
    ae.plot_histogram_fit()
    assert ae.attribute_center
    assert ae.attribute_sigma
    ae.report()

    ae_2 = eval(repr(ae))
    assert ae_2.parameter == {'limits': (100, 110)}


def test_Analysis_example_2():
    ae = AnalysisExampleAlgorithm_2()
    assert repr(ae) == 'AnalysisExampleAlgorithm_2(n_sample=100, seed=None)'
    ae.compute()
    assert len(ae.results) == 100
    ae.plot()
    ae.plot_2()
    ae.plot_histogram_fit()
    assert ae.attribute_center
    assert ae.attribute_sigma
    ae.report()


def test_pickling_Analysis():
    ae = AnalysisExampleAlgorithm_1(limits=(100, 110), meta={'comment': 'this is an example'})
    ae.compute(locdata=None)
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_analysis.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(ae, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            ae_pickled = pickle.load(file)
        assert ae_pickled.parameter == ae.parameter
        assert isinstance(ae_pickled.meta, metadata_analysis_pb2.AMetadata)
        assert ae_pickled.meta == ae.meta
